# frozen_string_literal: true

require "containers_cp/constants"
require "containers_cp/assets/helpers/resource_helper"

module ContainersCp::Assets
  module Containers
    module DaemonSets
      ASSET_NAME = "daemonsets"

      class Processor < ::ContainersCp::Processing::AssetProcessor
        include ::ContainersCp::Assets::ResourceHelper

        register source: DATA_SOURCE_K8S_AGENT, service: ASSET_SERVICE, asset: ASSET_NAME, format: FORMAT_JSON

        model ::KubernetesDaemonset

        asset_identifiers(:containers_cluster_id, :kubernetes_assigned_uid)

        copy_attributes_from_dso %w[name kubernetes_namespace_name kubernetes_cluster_name containers_cluster_id kubernetes_assigned_uid]

        set_callback :process_assets, :before, :clean_data

        def clean_data
          data.each do |dso|
            dso[:name] = dso[:metadata][:name]
            dso[:kubernetes_assigned_uid] = dso[:metadata][:uid]
            dso[:kubernetes_namespace_name] = dso[:metadata][:namespace]
            dso[:kubernetes_cluster_name] = @cluster_name
            dso[:containers_cluster_id] = @parent_cluster.id
          end
        end

        def process_asset(daemonset, dso)
          daemonset_metadata = dso[:metadata]

          # Find namespace and set kubernetes_namespace_id
          daemonset.kubernetes_namespace_id = load_asset_info(::KubernetesNamespace, daemonset.kubernetes_namespace_name).id

          # Populate account info
          if @parent_cluster.has_owner_account
            update_asset_account_info(daemonset)
          else
            logger.debug "No account info found on cluster: #{@cluster_name}, ID: #{@parent_cluster.id}"
            wavefront.counter("containers_cp.daemonset.no_account_info", 1, cluster: @cluster_name)
          end

          # Add Labels and Annotations
          process_labels_and_annotations(daemonset, daemonset_metadata)

          count_changed_assets(daemonset)
        end
      end
    end
  end
end

if Application.resolve("runtime").sandbox?
  def generate_daemonset_job(customer_id)
    generate_job customer_id, "daemonsets"
  end

  def process_daemonset(customer_id)
    process_payload ContainersCp::Assets::Containers::DaemonSets::Processor.new, customer_id, "daemonsets"
  end
end
