# frozen_string_literal: true

require "containers_cp/constants"
require "containers_cp/assets/helpers/resource_helper"

module ContainersCp::Assets
  module Containers
    module Deployments
      ASSET_NAME = "deployments"

      class Processor < ::ContainersCp::Processing::AssetProcessor
        include ::ContainersCp::Assets::ResourceHelper

        register source: DATA_SOURCE_K8S_AGENT, service: ASSET_SERVICE, asset: ASSET_NAME, format: FORMAT_JSON

        model ::KubernetesDeployment

        asset_identifiers(:containers_cluster_id, :kubernetes_assigned_uid)

        copy_attributes_from_dso %w[name kubernetes_namespace_name num_replicas kubernetes_cluster_name containers_cluster_id kubernetes_assigned_uid]

        set_callback :process_assets, :before, :clean_data

        def clean_data
          data.each do |dso|
            dso[:name] = dso[:metadata][:name]
            dso[:kubernetes_assigned_uid] = dso[:metadata][:uid]
            dso[:kubernetes_namespace_name] = dso[:metadata][:namespace]
            dso[:num_replicas] = dso[:spec][:replicas]
            dso[:kubernetes_cluster_name] = @cluster_name
            dso[:containers_cluster_id] = @parent_cluster.id
          end
        end

        def process_asset(deployment, dso)
          deployment_metadata = dso[:metadata]

          # Find namespace and set kubernetes_namespace_id
          deployment.kubernetes_namespace_id = load_asset_info(::KubernetesNamespace, deployment.kubernetes_namespace_name).id

          # Populate account info
          if @parent_cluster.has_owner_account
            update_asset_account_info(deployment)
          else
            logger.debug "No account info found on cluster: #{@cluster_name}, ID: #{@parent_cluster.id}"
            wavefront.counter("containers_cp.deployment.no_account_info", 1, cluster: @cluster_name)
          end

          # Add Labels and Annotations
          process_labels_and_annotations(deployment, deployment_metadata)

          count_changed_assets(deployment)
        end
      end
    end
  end
end

if Application.resolve("runtime").sandbox?
  def generate_deployment_job(customer_id)
    generate_job customer_id, "deployments"
  end

  def process_deployment(customer_id)
    process_payload ContainersCp::Assets::Containers::Deployments::Processor.new, customer_id, "deployments"
  end
end
