# frozen_string_literal: true

require "containers_cp/constants"
require "containers_cp/assets/helpers/resource_helper"

module ContainersCp::Assets
  module Containers
    module LimitRanges
      ASSET_NAME = "limitranges"

      class Processor < ::ContainersCp::Processing::AssetProcessor
        include ::ContainersCp::Assets::ResourceHelper

        register source: DATA_SOURCE_K8S_AGENT, service: ASSET_SERVICE, asset: ASSET_NAME, format: FORMAT_JSON

        model ::KubernetesNamespace

        asset_identifiers(:name, :containers_cluster_id)

        copy_attributes_from_dso %w[kubernetes_assigned_uid kubernetes_cluster_name]

        set_callback :process_assets, :before, :clean_data

        def clean_data
          data.each do |dso|
            dso[:name] = dso[:metadata][:namespace]
            # Create a new namespace and set kubernetes_assigned_uid as "uncollected" if it doesn't exist
            dso[:kubernetes_assigned_uid] = load_asset_info(::KubernetesNamespace, dso[:name]).kubernetes_assigned_uid
            dso[:kubernetes_cluster_name] = @cluster_name
            dso[:containers_cluster_id] = @parent_cluster.id
          end
        end

        def process_asset(namespace, dso)
          limit_range_spec = dso[:spec]

          limits = get_container_limits(limit_range_spec)
          if limits
            # Update kubernetes_namespace table columns (limit_range_mem_gb and limit_range_cpu) with corresponding values
            namespace.limit_range_mem_gb = parse_memory_gb(limits[:default][:memory])
            limit_range_cpu = parse_resource(limits[:default][:cpu])
            if limit_range_cpu
              if limit_range_cpu <= MAXIMUM_ALLOWED_CPU
                namespace.limit_range_cpu = limit_range_cpu
              else
                logger.warn "Namespace limit range CPU value over allowed DB limit: #{limits[:default][:cpu]} : #{limit_range_cpu}"
              end
            end
          end

          # Create a new entry for namespace and set kubernetes_assigned_uid as uncollected
          namespace.kubernetes_assigned_uid = "Uncollected" if namespace.kubernetes_assigned_uid.nil?
        end

        # find the limits entry for Container, ensure it contains default limits
        def get_container_limits(limit_range_spec)
          return nil unless limit_range_spec[:limits]
          limit_range_spec[:limits].each do |l|
            if l[:type] == "Container" && l[:default]
              return l
            end
          end
          nil
        end

        # This will avoid inadvertently deactivating namespaces since ResourceQuotas is part of kubernetes_namespace table
        def condensed?
          true
        end
      end
    end
  end
end

if Application.resolve("runtime").sandbox?
  def generate_limitrange_job(customer_id)
    generate_job customer_id, "limitranges"
  end

  def process_limitrange
    process_payload ContainersCp::Assets::Containers::LimitRanges::Processor.new, customer_id, "limitranges"
  end
end
