# frozen_string_literal: true

require "containers_cp/constants"
require "containers_cp/assets/helpers/resource_helper"
require "containers_cp/workers/asset_update_worker"

module ContainersCp::Assets
  module Containers
    module Nodes
      ASSET_NAME = "nodes"

      class Processor < ::ContainersCp::Processing::AssetProcessor
        include ::ContainersCp::Assets::ResourceHelper

        register source: DATA_SOURCE_K8S_AGENT, service: ASSET_SERVICE, asset: ASSET_NAME, format: FORMAT_JSON

        model ::KubernetesNode

        asset_identifiers(:name, :containers_cluster_id, :kubernetes_assigned_uid)

        copy_attributes_from_dso %w[kubernetes_cluster_name kubelet_version kubeproxy_version spec_provider_id kubernetes_assigned_uid]

        set_callback :process_assets, :before, :clean_data
        set_callback :process_assets, :before, :init_caches

        # Disable payload cache check for nodes, to support above timestamp update
        def support_caching?
          false
        end

        def clean_data
          data.each do |dso|
            dso[:name] = dso[:metadata][:name]
            dso[:kubernetes_assigned_uid] = dso[:metadata][:uid]
            dso[:kubernetes_cluster_name] = @cluster_name
            dso[:containers_cluster_id] = @parent_cluster.id
            if dso[:status][:nodeInfo]
              dso[:kubelet_version] = dso[:status][:nodeInfo][:kubeletVersion]
              dso[:kubeproxy_version] = dso[:status][:nodeInfo][:kubeProxyVersion]
            end
            dso[:spec_provider_id] = dso[:spec][:providerID]
          end
        end

        def process_asset(node, dso)
          node_status = dso[:status]
          node_metadata = dso[:metadata]
          node_spec = dso[:spec]

          if node_status[:allocatable]
            # Memory given in KiB.
            node.allocatable_mem_gb = parse_memory_gb(node_status[:allocatable][:memory])
            allocatable_cpu = parse_resource(node_status[:allocatable][:cpu])
            # sanity check for max value
            if allocatable_cpu
              if allocatable_cpu <= MAXIMUM_ALLOWED_CPU
                node.allocatable_cpu = allocatable_cpu
              else
                logger.warn "Container requested CPU value over allowed DB limit: #{node_status[:allocatable][:cpu]} : #{allocatable_cpu}"
              end
            end
          end

          update_instance_id(node)

          # update cluster ID and get account info from cluster if available
          # If account info not present, do the resolution process
          #  - this should happen for the first asset only, subsequent items will get it from the cluster
          if @parent_cluster.has_owner_account
            update_asset_account_info(node)
          else
            logger.debug "No account info for cluster: #{@cluster_name}, ID: #{@parent_cluster.id}, must resolve account."
            if resolve_account_from_provider(@parent_cluster, @customer.id, node_spec[:providerID])
              update_asset_account_info(node)
              trigger_asset_update_job
            else
              logger.warn("Could not resolve account for customer #{@customer.id} with providerID value: #{node_spec[:providerID]}")
              wavefront.counter("containers_cp.nodes.no_account_found.", 1, cluster: @cluster_name)
            end
          end
          process_labels_and_annotations(node, node_metadata)

          # ensure timestamp updated on the first asset we process
          if !@timestamp_updated
            node.updated_at = Time.now
            @timestamp_updated = true
          end
          count_changed_assets(node)
        end

        # build job context from parent cluster, submit job to sidekiq
        def trigger_asset_update_job
          job_args = {}
          job_args["customer_id"] = @customer.id
          job_args["parent_cluster_id"] = @parent_cluster.id
          job_args["cloud_provider"] = @parent_cluster.cloud_provider
          job_args["cloud_account_id"] = @parent_cluster.cloud_account_id
          job_args["aws_account_id"] = @parent_cluster.aws_account_id if @parent_cluster.aws_account_id
          job_args["gcp_compute_project_id"] = @parent_cluster.gcp_compute_project_id if @parent_cluster.gcp_compute_project_id
          job_args["azure_subscription_id"] = @parent_cluster.azure_subscription_id if @parent_cluster.azure_subscription_id
          # Schedule with a bit of delay to reduce chance of conflict with this cluster's data
          logger.debug "Triggering Account update job for customer #{@customer.id}, provider: #{@parent_cluster.cloud_provider}"
          ::ContainersCp::Workers::AssetUpdateWorker.perform_in(5.minutes, job_args)
        end

        def init_caches
          # cache instances to save DB lookups
          @cached_instances = {}
          # we need to set the updated_at timestamp on at least one node, so we'll use this to indicate this is done
          @timestamp_updated = false
        end

        def update_instance_id(node)
          provider_instance_id = extract_instance_id_from_provider(@parent_cluster.cloud_provider, node.spec_provider_id)
          if provider_instance_id
            case @parent_cluster.cloud_provider
            when PROVIDER_AWS
              instance_id = find_aws_instance(provider_instance_id, @parent_cluster.aws_account_id)
              if instance_id
                node.aws_instance_id = instance_id unless node.aws_instance_id == instance_id
              end
            when PROVIDER_GCP
              instance_id = find_gcp_compute_instance(provider_instance_id, @parent_cluster.gcp_compute_project_id)
              if instance_id
                node.gcp_compute_instance_id = instance_id unless node.gcp_compute_instance_id == instance_id
              end
            when PROVIDER_AZURE
              vm_id = find_azure_vm(provider_instance_id, @parent_cluster.azure_subscription_id)
              if vm_id
                node.azure_vm_id = vm_id unless node.azure_vm_id == vm_id
              end
            end
          end
        end

        def find_aws_instance(instance_id, aws_account_id)
          return nil unless instance_id

          unless @cached_instances.has_key? instance_id
            # We actually want to save `nil` because we don't want to lookup the same twice, even if it doesn't exist
            instance = ::AwsInstance
              .where(customer_id: @customer_id, aws_account_id: aws_account_id, instance_id: instance_id)
              .first
            @cached_instances[instance_id] = instance ? instance.id : nil
          end
          @cached_instances[instance_id]
        end

        def find_gcp_compute_instance(instance_id, gcp_id)
          return nil unless instance_id

          unless @cached_instances.has_key? instance_id
            # We actually want to save `nil` because we don't want to lookup the same twice, even if it doesn't exist
            instance = ::GcpComputeInstance
              .where(customer_id: @customer_id, gcp_compute_project_id: gcp_id, name: instance_id)
              .first
            @cached_instances[instance_id] = instance ? instance.id : nil
          end
          @cached_instances[instance_id]
        end

        def find_azure_vm(instance_id, azure_sub_id)
          return nil unless instance_id

          unless @cached_instances.has_key? instance_id
            # We actually want to save `nil` because we don't want to lookup the same twice, even if it doesn't exist
            instance = ::AzureVm
              .where(customer_id: @customer_id, azure_subscription_id: azure_sub_id, instance_id: instance_id)
              .first
            @cached_instances[instance_id] = instance ? instance.id : nil
          end
          @cached_instances[instance_id]
        end
      end
    end
  end
end

if Application.resolve("runtime").sandbox?
  def generate_node_job(customer_id)
    generate_job customer_id, "nodes"
  end

  def process_node(customer_id)
    process_payload ContainersCp::Assets::Containers::Nodes::Processor.new, customer_id, "nodes"
  end
end
