# frozen_string_literal: true

require "containers_cp/constants"
require "containers_cp/assets/helpers/resource_helper"

module ContainersCp::Assets
  module Containers
    module Pods
      ASSET_NAME = "pods"

      class Processor < ::ContainersCp::Processing::AssetProcessor
        include ::ContainersCp::Assets::ResourceHelper

        register source: DATA_SOURCE_K8S_AGENT, service: ASSET_SERVICE, asset: ASSET_NAME, format: FORMAT_JSON

        model ::KubernetesContainersImage

        asset_identifiers(:unique_container_image_key)

        set_callback :process_assets, :before, :preprocess_container_data
        set_callback :process_assets, :before, :load_caches
        set_callback :process_assets, :before, :clean_data

        # For pods, break out all of the containers into their own DSO object
        def preprocess_container_data
          new_data = []
          data.each do |dso|
            if dso[:spec] && dso[:spec][:containers]
              dso[:spec][:containers].each do |container|
                container_data = {}
                container_data[:metadata] = dso[:metadata]
                container_data[:container] = container
                new_data.append(container_data)
              end
            end
          end
          self.data = new_data unless new_data.blank?
        end

        def clean_data
          data.each do |dso|
            dso[:name] = dso[:container][:name]
            dso[:pod_name] = dso[:metadata][:name]
            dso[:kubernetes_namespace_name] = dso[:metadata][:namespace]

            # Recognized OwnerReference Types: Deployment, DaemonSet, ReplicaSet, StatefulSet, Job
            # Other OwnerReference Types: Nodes
            # None OwnerReference Types: any pod with empty or null ownerReferences
            load_owner_reference(dso[:metadata][:ownerReferences], dso)

            # Create/Set kubernetes_<asset>_id for Recognized OwnerReference Types: Deployment, DaemonSet, ReplicaSet, StatefulSet, Job
            create_owner_reference(dso)

            # Based on "cluster:namespace:workload_type:workload:container_image"
            load_unique_key(dso)
          end
        end

        def process_asset(container_image, dso)
          # check if we've already processed this unique key, skip if so
          if @unique_keys_processed.include?(dso[:unique_container_image_key])
            @metrics.count_assets_skipped
            throw :skip
          end
          container_image_metadata = dso[:metadata]
          container_image_container = dso[:container]
          container_request = container_image_container[:resources][:requests]
          container_limit = container_image_container[:resources][:limits]

          container_image.name = dso[:name]
          container_image.pod_name = dso[:pod_name]
          container_image.kubernetes_namespace_name = dso[:kubernetes_namespace_name]
          container_image.workload_type = dso[:workload_type]
          container_image.workload_name = dso[:workload_name]
          container_image.kubernetes_daemonset_id = dso[:kubernetes_daemonset_id]
          container_image.kubernetes_replicaset_id = dso[:kubernetes_replicaset_id]
          container_image.kubernetes_deployment_id = dso[:kubernetes_deployment_id]
          container_image.kubernetes_statefulset_id = dso[:kubernetes_statefulset_id]
          container_image.kubernetes_job_id = dso[:kubernetes_job_id]

          # Find namespace and set kubernetes_namespace_id
          container_image.kubernetes_namespace_id = load_asset_info(::KubernetesNamespace, container_image.kubernetes_namespace_name).id

          # Memory given in KiB.
          if container_request
            requested_cpu = parse_resource(container_request[:cpu]) if container_request[:cpu]
            # sanity check for max value
            if requested_cpu
              if requested_cpu <= MAXIMUM_ALLOWED_CPU
                container_image.requested_cpu = requested_cpu
              else
                logger.warn "Container requested CPU value over allowed DB limit: #{container_request[:cpu]} : #{requested_cpu}"
              end
            end
            container_image.requested_mem_gb = parse_memory_gb(container_request[:memory]) if container_request[:memory]
          end

          if container_limit
            limit_cpu = parse_resource(container_limit[:cpu]) if container_limit[:cpu]
            # sanity check for max value
            if limit_cpu
              if limit_cpu <= MAXIMUM_ALLOWED_CPU
                container_image.limit_cpu = limit_cpu
              else
                logger.warn "Container limit CPU value over allowed DB limit: #{container_limit[:cpu]} : #{limit_cpu}"
              end
            end
            container_image.limit_mem_gb = parse_memory_gb(container_limit[:memory]) if container_limit[:memory]
          end

          # Populate cluster and account info
          container_image.kubernetes_cluster_name = @cluster_name
          container_image.containers_cluster_id = @parent_cluster.id

          if @parent_cluster.has_owner_account
            update_asset_account_info(container_image)
          else
            logger.debug "No account info found on cluster: #{@cluster_name}, ID: #{@parent_cluster.id}"
            wavefront.counter("containers_cp.container_images.no_account_info", 1, cluster: @cluster_name)
          end

          # Add Labels and Annotations
          process_labels_and_annotations(container_image, container_image_metadata)

          # keep track of unique keys we've already processed
          @unique_keys_processed[dso[:unique_container_image_key]] = 1

          count_changed_assets(container_image)
        end

        def load_owner_reference(owner_references, dso)
          if owner_references
            name = owner_references.first[:name]
            type = owner_references.first[:kind].downcase
            dso[:workload_type] = if ["deployment", "daemonset", "replicaset", "statefulset", "job"].include? type
              type || ""
            else
              "other"
            end
            dso[:workload_name] = name || ""
          else
            dso[:workload_type] = "none"
            dso[:workload_name] = ""
          end
        end

        def load_unique_key(dso)
          # Unique key would be "cluster:namespace:workload_type:workload:container_image"
          dso[:unique_container_image_key] = @cluster_name +
              ":" + dso[:kubernetes_namespace_name] +
              ":" + dso[:workload_type] +
              (dso[:workload_name].empty? ? "" : ":" + dso[:workload_name]) +
              ":" + dso[:name]
        end

        def load_caches
          # track unique keys that we've already processed
          @unique_keys_processed = {}
          @metrics.start_cache_load_timer
          ::Multidb.use(@customer.read_shard) do
            load_assets(::KubernetesDeployment)
            load_assets(::KubernetesDaemonset)
            load_assets(::KubernetesReplicaset)
            load_assets(::KubernetesStatefulset)
            load_assets(::KubernetesJob)
          end
          @metrics.stop_cache_load_timer
        end

        def create_owner_reference(dso)

          case dso[:workload_type]
          when "deployment"
            owner_asset = load_asset_info(::KubernetesDeployment,
                                          dso[:workload_name],
                                          dso[:metadata][:ownerReferences].first[:uid],
                                          dso[:kubernetes_namespace_name])

            dso[:kubernetes_deployment_id] = owner_asset.id if owner_asset

            if dso[:kubernetes_deployment_id]
              @metrics.increment_counter "containers_cp.pods.deployment.id_set"
            else
              @metrics.increment_counter "containers_cp.pods.deployment.id_not_set"
            end
          when "daemonset"
            owner_asset = load_asset_info(::KubernetesDaemonset,
                                          dso[:workload_name],
                                          dso[:metadata][:ownerReferences].first[:uid],
                                          dso[:kubernetes_namespace_name])

            dso[:kubernetes_daemonset_id] = owner_asset.id if owner_asset

            if dso[:kubernetes_daemonset_id]
              @metrics.increment_counter "containers_cp.pods.daemonset.id_set"
            else
              @metrics.increment_counter "containers_cp.pods.daemonset.id_not_set"
            end
          when "replicaset"
            # Case 1: If ReplicaSet doesn't exist in DB, then will truncate workload_name and create a new ReplicaSet
            # Case 2: If ReplicaSet doesn't exist in DB (but ReplicaSet of truncated workload_name exists),
            # then will truncate workload_name and create a new ReplicaSet
            cache_key = [::KubernetesReplicaset.name, @customer.id, @parent_cluster.id, dso[:workload_name], dso[:kubernetes_namespace_name]]
            asset_fields = {
              customer_id: @customer.id,
              containers_cluster_id: @parent_cluster.id,
              name: dso[:workload_name],
              kubernetes_namespace_name: dso[:kubernetes_namespace_name]
            }
            unless asset_exist?(::KubernetesReplicaset, cache_key, asset_fields)
              dso[:workload_name] = trim_workload_name(dso[:workload_name])
            end

            replicaset = load_asset_info(::KubernetesReplicaset,
              dso[:workload_name],
              dso[:metadata][:ownerReferences].first[:uid],
              dso[:kubernetes_namespace_name])

            # Case 3: If ReplicaSet have a parent of Deployment and does exist in DB, update containers_image based on
            # Deployment
            if replicaset && replicaset.kubernetes_deployment_id
              dso[:kubernetes_deployment_id] = replicaset.kubernetes_deployment_id
              dso[:workload_type] = "deployment"
              dso[:workload_name] = replicaset.kubernetes_deployment_name || ""
              @metrics.increment_counter "containers_cp.pods.replicaset.deployment_id_set"
            # Case 4: If ReplicaSet doesn't have a parent of Deployment and NOT truncated workload_name exists in DB,
            # then will truncate workload_name and create a new ReplicaSet
            else
              dso[:workload_name] = trim_workload_name(dso[:workload_name])
              replicaset = load_asset_info(::KubernetesReplicaset,
                dso[:workload_name],
                dso[:metadata][:ownerReferences].first[:uid],
                dso[:kubernetes_namespace_name])

              @metrics.increment_counter "containers_cp.pods.replicaset.deployment_id_not_set"
            end

            if replicaset
              dso[:kubernetes_replicaset_id] = replicaset.id
              @metrics.increment_counter "containers_cp.pods.replicaset.id_set"
            else
              @metrics.increment_counter "containers_cp.pods.replicaset.id_not_set"
            end
          when "statefulset"
            owner_asset =  load_asset_info(::KubernetesStatefulset,
                                             dso[:workload_name],
                                             dso[:metadata][:ownerReferences].first[:uid],
                                             dso[:kubernetes_namespace_name])

            dso[:kubernetes_statefulset_id] = owner_asset.id if owner_asset

            if dso[:kubernetes_statefulset_id]
              @metrics.increment_counter "containers_cp.pods.statefulset.id_set"
            else
              @metrics.increment_counter "containers_cp.pods.statefulset.id_not_set"
            end
          when "job"
            # If Pod/Container is type Job, then truncate the ownerReference ID (last part) to get the correct name
            dso[:workload_name] = trim_workload_name(dso[:workload_name])

            owner_asset = load_asset_info(::KubernetesJob,
                                         dso[:workload_name],
                                         dso[:metadata][:ownerReferences].first[:uid],
                                         dso[:kubernetes_namespace_name])

            dso[:kubernetes_job_id] = owner_asset.id if owner_asset

            if dso[:kubernetes_job_id]
              @metrics.increment_counter "containers_cp.pods.job.id_set"
            else
              @metrics.increment_counter "containers_cp.pods.job.id_not_set"
            end
          end
        end

        # remove numeric suffix on workload name for jobs
        def trim_workload_name(name)
          parts = name.split("-")
          is_valid_hex?(parts.last) ? parts[0..-2].join("-") : parts.join("-")
        end

        def is_valid_hex?(s)
          # sanity check so we don't truncate things like "sidekiq-3a" or "allocation-scheduler-pr-226" - jobs and replicasets have 8-10 chars
          return false unless s.length > 4
          !s[/\H/]
        end

        # Overriding the base termination method so we can also terminate assets based on their parent assets
        def inactivate_terminated_assets
          termination_list = assets_for_termination
          return if termination_list.blank?
          logger.debug "Inactivating #{termination_list.length} assets..."
          inactivated_at = ::Time.now
          termination_list.each do |asset|
            # If parent asset (exception to "Job") is active, keep related containers images active
            parent_asset = nil
            case asset.workload_type
            when "daemonset"
              parent_asset = ::KubernetesDaemonset.find(asset.kubernetes_daemonset_id)
            when "replicaset"
              parent_asset = ::KubernetesReplicaset.find(asset.kubernetes_replicaset_id)
            when "deployment"
              parent_asset = ::KubernetesDeployment.find(asset.kubernetes_deployment_id)
            when "statefulset"
              parent_asset = ::KubernetesStatefulset.find(asset.kubernetes_statefulset_id)
            end

            if parent_asset&.is_active?
              logger.debug "Keep containers images active: #{asset.id}-#{asset.name}"
              @metrics.increment_counter "containers_cp.pods.keep_active"
              next
            end

            ::CloudHealth::Processing::Actions::TerminateAsset.call asset: asset,
                                                                    inactivated_at: inactivated_at,
                                                                    audit_log: audit_log,
                                                                    account: account

            # If parent asset is "Job" and part of termination list, explicitly call `TerminateAsset` on parent asset
            # We don't have other inactivation logic for "Job" because there is no other termination process for these
            if asset.workload_type == "job"
              job = ::KubernetesJob.find(asset.kubernetes_job_id)
              if job
                logger.debug "Inactivating related job asset: #{job.id}-#{job.name}"
                ::CloudHealth::Processing::Actions::TerminateAsset.call asset: job,
                                                                        inactivated_at: inactivated_at,
                                                                        audit_log: audit_log,
                                                                        account: account
              end
            end
          end
        end
      end
    end
  end
end

if Application.resolve("runtime").sandbox?
  def generate_pod_job(customer_id)
    generate_job customer_id, "pods"
  end

  def process_pod(customer_id)
    process_payload ContainersCp::Assets::Containers::Pods::Processor.new, customer_id, "pods"
  end
end
