# frozen_string_literal: true

require "containers_cp/constants"
require "containers_cp/assets/helpers/resource_helper"

module ContainersCp::Assets
  module Containers
    module ReplicaSets
      ASSET_NAME = "replicasets"

      class Processor < ::ContainersCp::Processing::AssetProcessor
        include ::ContainersCp::Assets::ResourceHelper

        register source: DATA_SOURCE_K8S_AGENT, service: ASSET_SERVICE, asset: ASSET_NAME, format: FORMAT_JSON

        model ::KubernetesReplicaset

        asset_identifiers(:containers_cluster_id, :kubernetes_assigned_uid)

        copy_attributes_from_dso %w[name kubernetes_namespace_name kubernetes_cluster_name containers_cluster_id kubernetes_assigned_uid]

        set_callback :process_assets, :before, :clean_data
        set_callback :process_assets, :before, :load_caches

        def clean_data
          data.each do |dso|
            dso[:name] = dso[:metadata][:name]
            dso[:kubernetes_assigned_uid] = dso[:metadata][:uid]
            dso[:kubernetes_namespace_name] = dso[:metadata][:namespace]
            dso[:kubernetes_cluster_name] = @cluster_name
            dso[:containers_cluster_id] = @parent_cluster.id
          end
        end

        def process_asset(replicaset, dso)
          replicaset_metadata = dso[:metadata]

          # Find namespace and set kubernetes_namespace_id
          replicaset.kubernetes_namespace_id = load_asset_info(::KubernetesNamespace, replicaset.kubernetes_namespace_name).id

          if replicaset_metadata[:ownerReferences] && replicaset_metadata[:ownerReferences].first[:kind] == "Deployment"
            replicaset.kubernetes_deployment_name = replicaset_metadata[:ownerReferences].first[:name]

            # Find deployment and set kubernetes_deployment_id
            original_deployment_id = replicaset.kubernetes_deployment_id
            deployment = load_asset_info(::KubernetesDeployment,
                                         replicaset.kubernetes_deployment_name,
                                         replicaset_metadata[:ownerReferences].first[:uid],
                                         dso[:kubernetes_namespace_name])

            replicaset.kubernetes_deployment_id = deployment.id if deployment

            # If deployment_id was added or changed, then update the corresponding container_images table with the deployment_id
            if original_deployment_id != replicaset.kubernetes_deployment_id
              containers_images = load_containers_image(dso[:kubernetes_namespace_name], dso[:name])
              update_containers_image(containers_images, replicaset)
            end

          end

          # Populate account info
          if @parent_cluster.has_owner_account
            update_asset_account_info(replicaset)
          else
            logger.debug "No account info found on cluster: #{@cluster_name}, ID: #{@parent_cluster.id}"
            wavefront.counter("containers_cp.replicasets.no_account_info", 1, cluster: @cluster_name)
          end

          # Add Labels and Annotations
          process_labels_and_annotations(replicaset, replicaset_metadata)

          count_changed_assets(replicaset)
        end

        def load_caches
          @metrics.start_cache_load_timer
          ::Multidb.use(@customer.read_shard) do
            load_assets(::KubernetesDeployment)
            load_assets(::KubernetesNamespace, false)
          end
          @metrics.stop_cache_load_timer
        end

        def update_containers_image(containers_images, replicaset)
          return if containers_images.nil?
          containers_images.each do |containers_image|
            containers_image.kubernetes_deployment_id = replicaset.kubernetes_deployment_id
            containers_image.workload_type = "deployment"
            containers_image.workload_name = replicaset.kubernetes_deployment_name || ""
            containers_image.unique_container_image_key = @cluster_name +
              ":" + containers_image.kubernetes_namespace_name +
              ":" + containers_image.workload_type +
              ":" + containers_image.workload_name +
              ":" + containers_image.name
            containers_image.save
          rescue ::ActiveRecord::RecordNotUnique
            # There can be a race condition when both ReplicaSet and Pod processor try to update same containers_image
            logger.debug "Duplicate update for containers_image: #{containers_image.unique_container_image_key}"
          end
        end
      end
    end
  end
end

if Application.resolve("runtime").sandbox?
  def generate_replicaset_job(customer_id)
    generate_job customer_id, "replicasets"
  end

  def process_replicaset(customer_id)
    process_payload ContainersCp::Assets::Containers::ReplicaSets::Processor.new, customer_id, "replicasets"
  end
end
