# frozen_string_literal: true

require "containers_cp/constants"
require "containers_cp/assets/helpers/resource_helper"

module ContainersCp::Assets
  module Containers
    module ResourceQuotas
      ASSET_NAME = "resourcequotas"

      class Processor < ::ContainersCp::Processing::AssetProcessor
        include ::ContainersCp::Assets::ResourceHelper

        register source: DATA_SOURCE_K8S_AGENT, service: ASSET_SERVICE, asset: ASSET_NAME, format: FORMAT_JSON

        model ::KubernetesNamespace

        asset_identifiers(:name, :containers_cluster_id)

        copy_attributes_from_dso %w[kubernetes_assigned_uid kubernetes_cluster_name]

        set_callback :process_assets, :before, :clean_data

        def clean_data
          data.each do |dso|
            dso[:name] = dso[:metadata][:namespace]
            # Create a new namespace and set kubernetes_assigned_uid as "uncollected" if it doesn't exist
            dso[:kubernetes_assigned_uid] = load_asset_info(::KubernetesNamespace, dso[:name]).kubernetes_assigned_uid
            dso[:kubernetes_cluster_name] = @cluster_name
            dso[:containers_cluster_id] = @parent_cluster.id
          end
        end

        def process_asset(namespace, dso)
          resource_quota_spec = dso[:status]

          # Update kubernetes_namespace table columns (resource_quota_cpu and resource_quota_mem_gb) with corresponding values
          if resource_quota_spec[:hard]
            namespace.resource_quota_mem_gb = parse_memory_gb(resource_quota_spec[:hard][:memory])
            resource_quota_cpu = parse_resource(resource_quota_spec[:hard][:cpu])
            # sanity check for max value
            if resource_quota_cpu
              if resource_quota_cpu <= MAXIMUM_ALLOWED_CPU
                namespace.resource_quota_cpu = resource_quota_cpu
              else
                logger.warn "Container requested CPU value over allowed DB limit: #{resource_quota_spec[:hard][:cpu]} : #{resource_quota_cpu}"
              end
            end
          end
        end

        # This will avoid inadvertently deactivating namespaces since ResourceQuotas is part of kubernetes_namespace table
        def condensed?
          true
        end
      end
    end
  end
end

if Application.resolve("runtime").sandbox?
  def generate_resourcequota_job(customer_id)
    generate_job customer_id, "resourcequotas"
  end

  def process_resourcequota(customer_id)
    process_payload ContainersCp::Assets::Containers::ResourceQuotas::Processor.new, customer_id, "resourcequotas"
  end
end
