# frozen_string_literal: true

require "karafka"

module ContainersCp
  module Consumers
    # This class will listen to the containers collection event kafka topic, validate the event contents,
    # and submit asset processing jobs to sidekiq.
    class ContainersFeedConsumer < ::Karafka::BaseConsumer
      include Import["logger", "alert", "wavefront", "cache.redis", "runtime", "settings"]
      # Kafka message format:
      # {"object_type"=>"NAMESPACES",
      #  "containers_cluster_id"=>"cluster-id",
      #  "customer_id"=>1,
      #  "s3_bucket"=>"cht-containers-replay-kubernetes",
      #  "s3_object_path"=>"1/2022/09/16/namespaces/virginia-1.dev.cloudhealthtech.com/23/1412485.json.gz",
      #   "kubernetes_agent_version" => "agent_v1",
      #  "kubernetes_cluster_uid"=>"78egh8e7h8e7h8"   <-- cluster_uid is optional, supported with new collector agent (May 2023)
      # }
      #
      CLUSTER_BLACKLIST_KEY = "cluster:blacklist"

      def consume

        if runtime.provider_aws? && !settings.wf_host.blank?
          # ensure wavefront thread is still alive, kill parent process if not
          unless Application["wavefront.thread"].alive?
            logger.error("Wavefront thread dead - flushing metrics and aborting consumer...")
            wavefront.flush
            Thread.main.exit
          end
        end

        begin
          batch_timer = ::Hitimes::Interval.now
          is_pushed = false
          params_batch.each do |msg_params|
            logger.debug { "ContainersFeedConsumer.consume, params: #{msg_params}" }
            validate_params!(msg_params)
            job_context = {"customer_id" => msg_params["customer_id"]}
            if msg_params["object_type"] == CLUSTER_REMOVE_OBJECT_TYPE || msg_params["object_type"] == CLUSTER_STATUS_OBJECT_TYPE
              handle_cluster_event(job_context, msg_params)
            else
              if is_excluded?(msg_params["customer_id"].to_s, msg_params["kubernetes_cluster_id"])
                logger.warn "ContainersFeedConsumer: skipping excluded cluster: #{msg_params["customer_id"]}-#{msg_params["kubernetes_cluster_id"]}"
                wavefront.counter("containers_cp.event.excluded_cluster", 1, cluster: msg_params["kubernetes_cluster_id"])
              else
                payload = create_payload(msg_params)
                if is_supported_asset?(payload)
                  unless is_pushed
                    logger.info "ContainersFeedConsumer.push_to_sidekiq, pushing ProcessingAssetsWorker job to sidekiq, partition: #{payload["partition"]}, offset: #{payload["offset"]}, create_time: #{payload["create_time"]}, batch_size: #{params_batch.count}"
                    is_pushed = true
                  end
                  push_to_sidekiq(job_context, payload)
                  wavefront.counter("containers_cp.event.dispatched", 1, asset_type: msg_params["object_type"])
                else
                  logger.warn "ContainersFeedConsumer: unsupported asset type: #{msg_params["object_type"]}"
                  wavefront.counter("containers_cp.event.unsupported_asset", 1, asset_type: msg_params["object_type"])
                end
              end
            end
          rescue ::Platform::CustomerNotFoundError
            logger.error "Customer not found on payload: #{msg_params}"
            wavefront.counter("containers_cp.dispatcher.event.missing_data", 1, property: "customer_id")
          rescue ::ContainersCp::AssestTypeNotFoundError
            logger.error "Object type not found on payload: #{msg_params}"
            wavefront.counter("containers_cp.dispatcher.event.missing_data", 1, property: "object_type")
          rescue ::ContainersCp::DataPayloadNotFoundError
            logger.error "S3 object path not found on payload: #{msg_params}"
            wavefront.counter("containers_cp.dispatcher.event.missing_data", 1, property: "s3_object_path")
          rescue ::ContainersCp::ClusterIdNotFoundError
            logger.error "K8S Cluster ID not found on payload: #{msg_params}"
            wavefront.counter("containers_cp.dispatcher.event.missing_data", 1, property: "kubernetes_cluster_id")
          rescue => ex
            logger.error "Bad message received on kafka topic. #{ex} - #{ex.message}"
            wavefront.counter("containers_cp.dispatcher.bad_event", 1)
          end
          batch_duration = batch_timer.stop * 1000 # Convert to milliseconds
          wavefront.dist("containers_cp.dispatcher.event.batch.duration", :m, batch_duration)
        rescue => ex
          logger.error "Error processing batch, skipping. #{ex} - #{ex.message}"
          wavefront.counter("containers_cp.dispatcher.batch_error", 1)
        end
      end

      def create_payload(params)
        payload = {}.merge(params)
        payload["data_source"] = DATA_SOURCE_K8S_AGENT
        payload["service"] = ASSET_SERVICE
        payload["format"] = FORMAT_JSON
        # cluster status and cluster remove require the same processor, so map these to the "cluster" asset type
        payload["asset"] = if params["object_type"] == CLUSTER_REMOVE_OBJECT_TYPE || params["object_type"] == CLUSTER_STATUS_OBJECT_TYPE
          CLUSTER_ASSET_TYPE
        else
          params["object_type"].downcase
        end
        payload
      end

      def is_supported_asset?(payload)
        processor_key = [payload["data_source"], payload["service"], payload["asset"], payload["format"]]
        return false if ::CloudHealth::Processing::Registry.instance[processor_key].nil?
        true
      end

      def push_to_sidekiq(context, payload)
        full_context = context.merge({"payload" => payload})
        ::ContainersCp::Workers::ProcessingAssetsWorker.perform_async(PROCESSOR_JOB_NAME, [full_context])
      end

      def validate_params!(msg_params)
        raise ::Platform::CustomerNotFoundError, msg_params if msg_params["customer_id"].to_i.zero?
        raise ::ContainersCp::AssestTypeNotFoundError unless msg_params["object_type"]
        raise ::ContainersCp::ClusterIdNotFoundError unless msg_params["kubernetes_cluster_id"]
        # cluster status and remove have no payload
        return if msg_params["object_type"].downcase == CLUSTER_REMOVE_OBJECT_TYPE || msg_params["object_type"].downcase == CLUSTER_STATUS_OBJECT_TYPE
        raise ::ContainersCp::DataPayloadNotFoundError unless msg_params["s3_object_path"]
      end

      def handle_cluster_event(job_context, params)
        job_context["action"] = ::ContainersCp::Workers::AssetUpdateWorker::ACTION_CLUSTER_REMOVE
        job_context["cluster_id"] = params["kubernetes_cluster_id"]
        logger.debug "Triggering Cluster Remove job for customer #{job_context["customer_id"]}, cluster ID: #{job_context["cluster_id"]}"
        ::ContainersCp::Workers::AssetUpdateWorker.perform_async(job_context)
      end

      def is_excluded?(cid, cluster)
        load_exclude_list
        if @exclude_list[cid]
          return @exclude_list[cid].include?(cluster)
        else
          return false
        end
      end

      # load list if we haven't yet, or if it's 10 min old
      def load_exclude_list
        return if @exclude_list && (Time.now.to_i - @exclude_list_updated) < 10.minutes
        cached_list = redis.from_hash(CLUSTER_BLACKLIST_KEY).with_indifferent_access
        @exclude_list = {}
        total_clusters = 0
        cached_list.each do |cid, v|
          clusters = v.split(",")
          total_clusters += clusters.length
          @exclude_list[cid] = clusters unless clusters.empty?
        end
        logger.info "loaded #{@exclude_list.size} customers into cluster exclude list, total clusters: #{total_clusters}"
        @exclude_list_updated = Time.now.to_i
      end
    end
  end
end
