require "hitimes"

module ContainersCp
  module Metrics
    # Wrapper that implements the statsd methods, but logs metrics using wavefront SDK
    # takes a prefix to add to the metric name (for metrics from the framework, where this is used)
    class StatsdWfWrapper
      def initialize(wavefront, prefix = "containers_cp")
        @wavefront = wavefront
        @prefix = prefix + "."
      end

      def increment(path, by: 1, tags: nil, sample_rate: 1)
        sample(sample_rate) do
          @wavefront.gauge(@prefix + path, by, tags_to_hash(tags))
        end
      end

      def gauge(path, value, tags: nil, sample_rate: 1)
        sample(sample_rate) do
          @wavefront.gauge(@prefix + path, value, tags_to_hash(tags))
        end
      end

      def time(path, tags: nil, sample_rate: 1)
        timer = ::Hitimes::Interval.now
        yield
        elapsed = timer.stop * 1000 # Convert to milliseconds
        timing(@prefix + path, elapsed, tags: tags, sample_rate: sample_rate)
      end

      def timing(path, time_val, tags: nil, sample_rate: 1)
        sample(sample_rate) do
          @wavefront.dist(@prefix + path, :m, time_val, tags_to_hash(tags))
        end
      end

      def histogram(path, value, args = {})
        tags = tags_to_hash(args[:tags])
        @wavefront.dist(@prefix + path, :m, value, tags)
      end

      def count(path, by, args: {})
        tags = tags_to_hash(args[:tags])
        @wavefront.counter(@prefix + path, by, tags)
      end

      # not supported, just yield with this class to log metrics
      def batch
        yield self
      end

      private

      def sample(sample_rate)
        if sample_rate != 1 && rand > sample_rate
          return
        end

        yield
      end

      def tags_to_hash(tags)
        return if tags.nil?
        tag_hash = {}
        tags.each do |t|
          tag_val = t.split(":")
          tag_hash[tag_val[0]] = tag_val[1]
        end
        tag_hash
      end
    end
  end
end
