module ContainersCp::Processing
  module AccountResolverHelper
    # Look up account info based on the cloud provider, by first extracting both the provider
    # and the relevant ID from the spec.providerId field.
    # Update the cluster with this provider and account information.
    #
    # NOTE: observed providerIDs not currently supported:
    # vsphere://42237e92-8889-5987-0c9a-f4df3c6e08f3
    # ibm://4183e714347c43a79451c9b58ed0e5fd///c5fbel3d03q1nfiuf5bg/kube-c5fbel3d03q1nfiuf5bg-devussouth-default-00003ac4
    #
    # 
    @logger = Application["logger"]
    
    def resolve_account_from_provider(cluster, customer_id, provider_id)
      return false if provider_id.nil?
      @logger.debug "resolve_account_from_provider provider_id:  #{provider_id}"
      if provider_id.start_with?("aws")
        instance_id = extract_account_id_from_provider(PROVIDER_AWS, provider_id)
        aws_account_id, owner_id = aws_instance_lookup(customer_id, instance_id)
        return false unless aws_account_id
        @logger.debug "AWS provider, found aws_account_id: #{aws_account_id} owner_id: #{owner_id}"
        cluster.aws_account_id = aws_account_id
        cluster.cloud_account_id = owner_id
        cluster.cloud_provider = PROVIDER_AWS
        cluster.auto_assigned = 1
        cluster.save
        true
      elsif provider_id.start_with?("gce")
        gcp_project_name = extract_account_id_from_provider(PROVIDER_GCP, provider_id)
        gcp_compute_project_id = gcp_compute_project_lookup(customer_id, gcp_project_name)
        return false unless gcp_compute_project_id
        @logger.debug "GCP Provider, found gcp_compute_project_id: #{gcp_compute_project_id} for project_name: #{gcp_project_name}"
        cluster.gcp_compute_project_id = gcp_compute_project_id
        cluster.cloud_account_id = gcp_project_name
        cluster.cloud_provider = PROVIDER_GCP
        cluster.auto_assigned = 1
        cluster.save
        true
      elsif provider_id.start_with?("azure")
        azure_id = extract_account_id_from_provider(PROVIDER_AZURE, provider_id)
        azure_subscription_id = azure_subscription_lookup(customer_id, azure_id)
        return false unless azure_subscription_id
        @logger.debug "Azure Provider, found azure_subscription_id: #{azure_subscription_id} for azure_id: #{azure_id}"
        cluster.azure_subscription_id = azure_subscription_id
        cluster.cloud_account_id = azure_id
        cluster.cloud_provider = PROVIDER_AZURE
        cluster.auto_assigned = 1
        cluster.save
        true
      else
        @logger.warn "Unknown provider on providerId: #{provider_id}"
        false
      end
    end

    # Extract cloud provider and relevant ID from the spec.providerId property. These have formats as followed for
    # each cloud provider:
    # AWS:  “aws:///<region>/<ec2_instance_id>”  ex: "aws:///us-east-1b/i-08d3f14dc74cac8c4"
    # GCP: “gce://<project_id>/<region>/<vm_instance>” ex: "gce://team-sudo/us-central1-a/gke-liveliness-check-080122-pool-1-827619bf-79kl""
    # Azure:  “azure:///subscriptions/<subscription_id>/resourceGroups/<resource_group_id>/providers/Microsoft.Compute/virtualMachineScaleSets/<vm_scaleset_name>/virtualMachines/<vm_number>"
    #    ex:  "azure:///subscriptions/a827c951-9c3a-4c0b-aed0-542c8b5ecc03/resourceGroups/mc_gmudumbai_group_acr_aks_test_1_eastus/providers/Microsoft.Compute/virtualMachineScaleSets/aks-agentpool-30548465-vmss/virtualMachines/0"
    def extract_account_id_from_provider(provider, provider_id)
      return nil if provider_id.nil?
      begin
        case provider
        when PROVIDER_AWS
          # instance ID after  last "/"
          idx = provider_id.rindex("/") + 1
          aws_instance_id = provider_id[idx..-1]
          @logger.debug "Extracted AWS instance ID: #{aws_instance_id} from: #{provider_id}"
          aws_instance_id
        when PROVIDER_GCP
          # extract between "gce://" and next "/"
          gidx1 = provider_id.index("://") + 3
          gidx2 = provider_id.index("/", gidx1) - 1
          gcp_compute_project_id = provider_id[gidx1..gidx2]
          @logger.debug "Extracted GCP project ID: #{gcp_compute_project_id} from: #{provider_id}"
          gcp_compute_project_id
        when PROVIDER_AZURE
          # extract between "subscriptions/" and next "/"
          aidx1 = provider_id.index("subscriptions/") + "subscriptions/".length
          aidx2 = provider_id.index("/", aidx1) - 1
          azure_sub_id = provider_id[aidx1..aidx2]
          @logger.debug "Extracted Azure subscription ID: #{azure_sub_id} from: #{provider_id}"
          azure_sub_id
        end
      rescue => ex
        @logger.warn "Exception parsing account from provider ID: #{provider_id} for provider: #{provider}"
        nil
      end
    end

    # Extract the instance/Virtual Machine ID from the "spec.ProviderId" property
    # AWS and GCP both have this as the last item, for Azure this is the entire value starting with "/subscriptions"
    # AWS:  “aws:///<region>/<ec2_instance_id>”  ex: "aws:///us-east-1b/i-08d3f14dc74cac8c4"
    # GCP: “gce://<project_id>/<region>/<vm_instance>” ex: "gce://team-sudo/us-central1-a/gke-liveliness-check-080122-pool-1-827619bf-79kl""
    # Azure:  “azure:///subscriptions/<subscription_id>/resourceGroups/<resource_group_id>/providers/Microsoft.Compute/virtualMachineScaleSets/<vm_scaleset_name>/virtualMachines/<vm_number>"
    #    ex:  "azure:///subscriptions/a827c951-9c3a-4c0b-aed0-542c8b5ecc03/resourceGroups/mc_gmudumbai_group_acr_aks_test_1_eastus/providers/Microsoft.Compute/virtualMachineScaleSets/aks-agentpool-30548465-vmss/virtualMachines/0"
    def extract_instance_id_from_provider(provider, provider_id)
      return nil if provider_id.nil?
      begin
        case provider
        when PROVIDER_AWS
          # instance ID after last "/"
          idx = provider_id.rindex("/") + 1
          provider_id[idx..-1]
        when PROVIDER_GCP
          # instance ID after last "/"  (same as AWS)
          idx = provider_id.rindex("/") + 1
          provider_id[idx..-1]
        when PROVIDER_AZURE
          # extract starting with "/subscriptions" to end of string
          idx = provider_id.index("/subscriptions")
          provider_id[idx..-1]
        end
      rescue => ex
        @logger.warn "Exception parsing instance from provider ID: #{provider_id} for provider: #{provider}"
        nil
      end
    end

    # look up the instance and associated account to get both the account ID and AWS owner ID
    def aws_instance_lookup(customer_id, instance_id)
      ::Multidb.use(@customer.read_shard) do
        instance = ::AwsInstance
          .includes(:aws_account)
          .where(
            customer_id: customer_id,
            instance_id: instance_id
          ).first
        @logger.warn "No instance found with id: #{instance_id}" unless instance
        return unless instance
        @logger.debug "aws_instance_lookup, found instance #{instance_id}, ID: #{instance.aws_account_id}, owner: #{instance.aws_account.owner_id}"
        return instance.aws_account_id, instance.aws_account.owner_id
      end
    end

    # look up GCP compute project by name
    def gcp_compute_project_lookup(customer_id, gcp_compute_project_name)
      ::Multidb.use(@customer.read_shard) do
        gcp_project = ::GcpComputeProject.where(customer_id: customer_id, name: gcp_compute_project_name).first
        @logger.warn "No gcp_project found with name: #{gcp_compute_project_name}" unless gcp_project
        return unless gcp_project
        @logger.debug "gcp_compute_project_lookup, found project ID #{gcp_project.id} for name: #{gcp_compute_project_name}"
        gcp_project.id
      end
    end

    # look up Azure subscription by azure_id
    def azure_subscription_lookup(customer_id, azure_id)
      ::Multidb.use(@customer.read_shard) do
        subscription = ::AzureSubscription.where(customer_id: customer_id, azure_id: azure_id).first
        @logger.warn "No azure subscription found with azure_id: #{azure_id}" unless subscription
        return unless subscription
        @logger.debug "azure_subscription_lookup, found subscription ID #{subscription.id} for azure_id: #{azure_id}"
        subscription.id
      end
    end
  end
end
