require_relative "account_resolver_helper"
require_relative "common"
require_relative "tag_update_helper"

module ContainersCp::Processing
  class AssetProcessor < ::CloudHealth::Processing::AssetProcessor
    include Import["logger", "alert", "wavefront", "cache.redis", "runtime"]
    include ::ContainersCp::Processing::AccountResolverHelper
    include ::ContainersCp::Processing::Common

    set_callback :prepare, :before, :init_data
    set_callback :process_assets, :after, :record_job_metrics

    def init_data
      @metrics = ContainersCp::Processing::MetricsHelper.new(wavefront, logger, cloud_asset)
      @audit_log = ::CloudHealth::Processing::AuditLog.new(
        @customer_id,
        account.id,
        service: self.class.cloud_service,
        asset: self.class.cloud_asset
      )

      # Spend verification jobs are special jobs sent from the scheduler when an asset has been
      # disabled for a while. Processing checks if any asset changes happened during this
      # special job and notifies the owning team.
      @spend_verification_job = false
      @spend_verification_job_alerted = false

      # Cache for saving DB lookups
      @kubernetes_cache = {}

      # For increment_dup_create_count in `cht-framework-cp`
      @dup_create_crns = []
    end

    def prepare
      # look up the parent cluster based on the containers_cluster_id on the incoming event
      # We need this to get account info, as well as scoping the asset queries
      load_cluster_info
      if !@parent_cluster.is_active
        logger.warn "inactive cluster! #{@parent_cluster.cluster_name}, rejecting feed"
        tags = {"cluster" => @parent_cluster.cluster_name}
        wavefront.counter("containers-cp.processor.inactive_cluster", 1, tags)
        return false
      end

      @lookup_attributes = { customer_id: @customer_id, containers_cluster_id: @parent_cluster.id }
      @assets = load_assets_cache
      logger.info "loaded #{klass_model.name} asset cache with #{@assets.size} assets for cluster: #{@parent_cluster.id}-#{@cluster_name} @ids: #{@ids.size}"

      label_table_name = klass_model.label_model.table_name
      label_ids = ::Multidb.use(@customer.read_shard) do
        base_asset_query.joins(label_table_name.to_sym).where(id: @ids.to_a).pluck(label_table_name + ".id").uniq
      end
      annotation_table_name = klass_model.annotation_model.table_name
      annotation_ids = ::Multidb.use(@customer.read_shard) do
        base_asset_query.joins(annotation_table_name.to_sym).where(id: @ids.to_a).pluck(annotation_table_name + ".id").uniq
      end

      @tag_update_helper = ContainersCp::Processing::TagUpdateHelper.new(
        @customer,
        klass_model,
        label_ids,
        annotation_ids,
        logger
      )

      true
    end

    def load_payload
      logger.debug "asset_processor.load_payload from path: #{@options[:payload][:path]}"
      @options[:payload][:path] ? fetch_payload : fetch_remote_payload
      # Initialize payload hash cache. This is used in `cht-framework-cp`.
      # It is stored in redis after a successful processing run.
      calculate_cache(@options[:payload][:data])
      super
    end

    def process_labels_and_annotations(asset, metadata)
      @tag_update_helper.update_asset_tags(asset, metadata["labels"], metadata["annotations"])
    end

    def create_parent_asset(parent_klass, required_fields, uid = nil)
      asset = parent_klass.new
      asset.attributes = required_fields
      if uid
        asset[:kubernetes_assigned_uid] = uid
        asset[:containers_cluster_id] = @parent_cluster.id
      end
      ::CloudHealth::Processing::Actions::CreateAsset.call asset: asset
      @metrics.count_parent_asset_created
      asset
    rescue ::ActiveRecord::RecordNotUnique
      # There can be a case where we need resurrect inactive parent assets
      @metrics.count_duplicate_created_asset
      logger.debug "Duplicate parent creation attempt for #{parent_klass} with these fields: #{required_fields}"
      parent_klass.unscoped.where(required_fields).first
    end

    def asset_exist?(model_class, cache_key, asset_fields)
      parent_asset = @kubernetes_cache[cache_key]
      if parent_asset.nil?
        # Save parent asset to cache even if it is nil, so we don't keep querying for it
        ::Multidb.use(@customer.read_shard) do
          parent_asset = @kubernetes_cache[cache_key] = model_class.active.where(asset_fields).first
        end
        @metrics.count_db_lookup
        # If active asset does not exist in AssetDB, return false
        return false if parent_asset.nil?
      else
        @metrics.count_asset_cache_hit
      end
      true
    end

    def load_asset_info(model_class, name, uid = "Uncollected", namespace_name = nil)
      cache_key = [model_class.name, @customer.id, @cluster_name, name]
      asset_fields = {
        customer_id: @customer.id,
        containers_cluster_id: @parent_cluster.id,
        name: name,
      }
      if model_class.column_names.include?("kubernetes_namespace_name") && namespace_name
        cache_key.append(namespace_name)
        asset_fields[:kubernetes_namespace_name] = namespace_name
      end
      return @kubernetes_cache[cache_key] if asset_exist?(model_class, cache_key, asset_fields)

      # We need a valid UID to create any related asset, except Namespaces
      if uid != "Uncollected" || model_class == ::KubernetesNamespace
        parent_asset = create_parent_asset(model_class, asset_fields, uid)
        @kubernetes_cache[cache_key] = parent_asset
        return parent_asset
      end
      nil
    end

    def load_containers_image(namespace_name, workload_name)
      containers_image = nil
      @metrics.count_db_lookup
      ::Multidb.use(@customer.data_shard) do
        containers_image = ::KubernetesContainersImage.where(customer_id: @customer.id,
          containers_cluster_id: @parent_cluster.id,
          kubernetes_namespace_name: namespace_name,
          workload_name: workload_name)
      end
      containers_image
    end

    private

    def cache_key_prefix
      [
        @customer.id,
        @options[:payload][:kubernetes_cluster_id],
        @options[:payload][:asset]
      ].join(":")
    end

    def calculate_cache(payload, key_suffix: nil)
      key = cache_key_prefix
      key = "#{cache_key_prefix}:#{key_suffix}" unless key_suffix.blank?
      @options[:cache] = {
        key: key,
        md5: payload_md5(payload)
      }
      @options[:ts] = Time.now.to_i
    end

    def payload_md5(json_payload)
      ::Digest::MD5.hexdigest(::JSON.fast_generate(json_payload))
    end

    def fetch_remote_payload
      if runtime.provider_gcp?
        fetch_gcs_payload
      else
        fetch_s3_payload
      end

    end

    def fetch_s3_payload
      payload_json = ::JSON.parse s3_access.get_uncompressed(@options[:payload][:s3_bucket], @options[:payload][:s3_object_path])
      @options[:payload][:data] = payload_json["items"]
    end

    def fetch_gcs_payload
      logger.debug "fetch_gcs_payload: fetching payload from gcs bucket: #{@options[:payload][:s3_bucket]}, path: #{@options[:payload][:s3_object_path]}"
      payload_json = ::JSON.parse gcs_access.get_uncompressed(@options[:payload][:s3_bucket], @options[:payload][:s3_object_path])
      logger.info "fetch_gcs_payload payload retrieved, contains #{payload_json["items"].size} items"
      @options[:payload][:data] = payload_json["items"]
    end

    def fetch_payload
      payload_json = ::JSON.parse Zlib::GzipReader.open(@options[:payload][:path]) { |gzip| gzip.read }
      @options[:payload][:data] = payload_json["items"]
    end

    # This method will now use the "kubernetes_cluster_uid" in the lookup if present in the payload. If we don't find
    # the cluster, try with the cluster name and DEFAULT_CLUSTER_UID. If this is found, update the UID on that cluster.
    # If not found, create a new cluster with the UID
    # We are also now receiving the "kubernetes_agent_version" on the incoming payload, so save this as well, and/or
    # update the existing cluster if this value has changed
    def load_cluster_info
      @cluster_name = @options[:payload][:kubernetes_cluster_id]
      has_uid =  @options[:payload][:kubernetes_cluster_uid]
      cluster_uid = has_uid ? @options[:payload][:kubernetes_cluster_uid] : DEFAULT_CLUSTER_UID
      logger.debug "load_cluster_info, @customer.id: #{@customer.id}, cluster_name: #{@cluster_name}, cluster_uid: #{cluster_uid}, agent_version:#{@options[:payload][:kubernetes_agent_version]}"
      ::Multidb.use(@customer.data_shard) do
        @parent_cluster = lookup_cluster(@customer.id, @cluster_name, cluster_uid)
        logger.debug "Found cluster ID #{@parent_cluster.id} - #{@cluster_name} with UID: #{cluster_uid}, agent_version: #{@parent_cluster.kubernetes_agent_version}" if @parent_cluster
        if @parent_cluster.nil? && has_uid
          @parent_cluster = lookup_cluster(@customer.id, @cluster_name, DEFAULT_CLUSTER_UID)
          # if found, update the cluster
          if @parent_cluster
            logger.debug "Found cluster ID #{@parent_cluster.id} - #{@cluster_name} without UID, updating UID to: #{cluster_uid}"
            @parent_cluster.uid = cluster_uid
            @parent_cluster.kubernetes_agent_version = @options[:payload][:kubernetes_agent_version]
            @parent_cluster.save
          end
        end
        if @parent_cluster.nil?
          cluster_fields = {
            customer_id: @customer.id,
            cluster_name: @cluster_name,
            cluster_type: "KUBERNETES",
            kubernetes_agent_version: @options[:payload][:kubernetes_agent_version],
            uid: cluster_uid
          }
          @parent_cluster = create_parent_asset(::ContainersCluster, cluster_fields)
        else
          # Check the version, save if it's changed
          if @options[:payload][:kubernetes_agent_version] &&
              @options[:payload][:kubernetes_agent_version] != @parent_cluster.kubernetes_agent_version
            logger.debug "Updating kubernetes_agent_version on cluster to: #{@options[:payload][:kubernetes_agent_version]}"
            @parent_cluster.kubernetes_agent_version = @options[:payload][:kubernetes_agent_version]
            @parent_cluster.save
          end
        end
      end
    end

    def lookup_cluster(customer_id, cluster_name, cluster_uid)
      ::ContainersCluster.active.where(customer_id: customer_id,
                                       cluster_name: cluster_name,
                                       uid: cluster_uid).first
    end

    # called from processor "load_caches" method if needed
    # ex:
    #         def load_caches
    #           @metrics.start_cache_load_timer
    #           ::Multidb.use(@customer.read_shard) do
    #             load_assets(::KubernetesDeployment)
    #           end
    #           @metrics.stop_cache_load_timer
    #         end
    # For assets that have a kubernetes_namespace_name, use this as part of the cache key
    def load_assets(model_class, has_namespace_name = true)
      model_class.active.where(customer_id: @customer_id, containers_cluster_id: @parent_cluster.id).each do |asset|
        cache_key = has_namespace_name ? [model_class.name, @customer.id, @parent_cluster.id, asset.name, asset.kubernetes_namespace_name]
                                       : [model_class.name, @customer.id, @parent_cluster.id, asset.name]
        @kubernetes_cache[cache_key] = asset
      end
    end

    # this method will update the appropriate columns with provider and account ID info from the related cluster
    # This assumes the asset will have the columns:
    # cloud_provider_id
    # cloud_account_id
    # aws_account_id
    # gcp_compute_project_id
    # azure_subscription_id
    def update_asset_account_info(asset)
      return unless @parent_cluster
      asset.cloud_provider_id = @parent_cluster.cloud_provider
      asset.cloud_account_id = @parent_cluster.cloud_account_id
      asset.aws_account_id = @parent_cluster.aws_account_id
      asset.gcp_compute_project_id = @parent_cluster.gcp_compute_project_id
      asset.azure_subscription_id = @parent_cluster.azure_subscription_id
    end

    def count_changed_assets(asset)
      if asset.changed?
        @metrics.count_asset_changed
      else
        @metrics.count_asset_unchanged
      end
    end

    def record_job_metrics
      @metrics.count_assets_processed(data.size)
      @metrics.record_metrics(@parent_cluster.cloud_provider)
    end

    def save_asset(asset)
      if asset.persisted?
        return :no_change unless asset.changed?
        begin
          ::CloudHealth::Processing::Actions::UpdateAsset.call asset: asset,
            audit_log: audit_log,
            account: account
          check_spend_verification_job
          :updated
        rescue => e
          # If there's a duplicate key
          # return :conflict, log and stat it.
          if ex_includes_dup_key(e)
            logger.warn "Duplicate entry found for asset being updated. "\
                         "ID: #{asset.id}, customer_id: #{@customer_id}, cluster: #{@parent_cluster.id}-#{@cluster_name} "\
                         "asset type: #{self.class.cloud_asset}"

            tags = {"asset_type" => asset.class.name}
            wavefront.counter("containers-cp.processor.update.duplicate_key", 1, tags)
            :conflict
          else
            raise e
          end
        end
      else
        begin
          ::CloudHealth::Processing::Actions::CreateAsset.call asset:     asset,
            audit_log: audit_log,
            account:   account
          :new
        rescue => e
          # If there's a duplicate key
          # return :conflict, log and stat it.
          if ex_includes_dup_key(e)
            asset_name = asset.respond_to?(:name) ? asset.name : "n/a"
            asset_ns_name = asset.respond_to?(:kubernetes_namespace_name) ? asset.kubernetes_namespace_name : "n/a"
            logger.warn "Duplicate entry found for asset being created. "\
                         "customer_id: #{@customer_id}, cluster: #{@parent_cluster.id}-#{@cluster_name} "\
                         "asset type: #{self.class.cloud_asset} name: #{asset_name}, ns_name: #{asset_ns_name}, crn: #{asset.crn}"

            tags = {"asset_type" => asset.class.name}
            wavefront.counter("containers-cp.processor.create.duplicate_key", 1, tags)
            :conflict
          else
            raise e
          end
        end
      end
    ensure
      @ids.delete asset.id
    end
  end
end
