require "hitimes"

module ContainersCp::Processing
  class MetricsHelper
    # This class maintains some counters to be used for processing, which allows us to tally up some counts, and
    # log them when processing is done, so we're not pounding wavefront too heavily during the job

    def initialize(wavefront, logger, asset_type)
      @wavefront = wavefront
      @logger = logger
      @asset_type = asset_type

      @logger.debug("Initialize MetricsHelper with asset_type: #{asset_type}")

      # counters
      @asset_db_lookup = 0
      @asset_cache_hit = 0
      @assets_processed = 0
      @assets_skipped = 0
      @asset_changed = 0
      @asset_unchanged = 0
      @parent_asset_created = 0
      @duplicate_created_asset = 0

      # allow for processors to define new counters
      @custom_counters = {}

      # timers
      @job_duration = 0
      @cache_load_duration = 0
      @tag_processing_duration = 0
      @asset_processing_duration = 0
      # start the job timer now, stop when we record metrics
      @job_timer = ::Hitimes::Interval.now
    end

    # counters
    def count_db_lookup(ct = 1)
      @asset_db_lookup += ct
    end

    def count_asset_cache_hit(ct = 1)
      @asset_cache_hit += ct
    end

    def count_assets_processed(ct = 1)
      @assets_processed += ct
    end

    def count_assets_skipped(ct = 1)
      @assets_skipped += ct
    end

    def count_parent_asset_created(ct = 1)
      @parent_asset_created += ct
    end

    def count_duplicate_created_asset(ct = 1)
      @duplicate_created_asset += ct
    end

    def count_asset_changed(ct = 1)
      @asset_changed += ct
    end

    def count_asset_unchanged(ct = 1)
      @asset_unchanged += ct
    end

    def increment_counter(metric_name)
      count = @custom_counters[metric_name]
      count = count ? count + 1 : 1
      @custom_counters[metric_name] = count
    end

    # timers
    def start_cache_load_timer
      @cache_load_timer = ::Hitimes::Interval.now
    end

    def stop_cache_load_timer
      @cache_load_duration = @cache_load_timer.stop * 1000 if @cache_load_timer
    end

    def start_asset_processing_timer
      @asset_processing_timer = ::Hitimes::Interval.now
    end

    def stop_asset_processing_timer
      @asset_processing_duration = @asset_processing_timer.stop * 1000 if @asset_processing_timer
    end

    # Record the job metris with WF. Pass provider in here since we don't have this when initialized
    def record_metrics(provider)
      @job_duration = @job_timer.stop * 1000 # Convert to milliseconds
      tags = {"asset_type" => @asset_type, "cloud_provider" => provider}
      if @wavefront
        # count this as one job (won't need this if we can get processor job metrics)
        @wavefront.counter("containers_cp.processing.job.count", 1, tags)
        @wavefront.counter("containers_cp.processing.asset.db.lookup", @asset_db_lookup, tags) if @asset_db_lookup > 0
        @wavefront.counter("containers_cp.processing.asset.cache.hit", @asset_cache_hit, tags) if @asset_cache_hit > 0
        @wavefront.counter("containers_cp.processing.assets.processed", @assets_processed, tags) if @assets_processed > 0
        @wavefront.counter("containers_cp.processing.assets.skipped", @assets_skipped, tags) if @assets_skipped > 0
        @wavefront.counter("containers_cp.processing.parent.asset.created", @parent_asset_created, tags) if @parent_asset_created > 0
        @wavefront.counter("containers_cp.processing.duplicate.created.asset", @duplicate_created_asset, tags) if @duplicate_created_asset > 0
        @wavefront.counter("containers_cp.processing.assets_changed", @asset_changed, tags) if @asset_changed > 0
        @wavefront.counter("containers_cp.processing.assets_unchanged", @asset_unchanged, tags) if @asset_unchanged > 0
        @wavefront.dist("containers_cp.processing.job.runtime", :m, @job_duration, tags) if @job_duration > 0
        @wavefront.dist("containers_cp.processing.cache.load.time", :m, @cache_load_duration, tags) if @cache_load_duration > 0
        @wavefront.dist("containers_cp.processing.asset.processing.time", :m, @asset_processing_duration, tags) if @asset_processing_duration > 0
        if @job_duration > 60000
          @logger.info "Processing Job Metrics - counters: asset_type: #{@asset_type}, assets.processed: #{@assets_processed}, assets_skipped: #{@assets_skipped}, assets_changed: #{@asset_changed}, asset_unchanged: #{@asset_unchanged}, asset.db.lookup: #{@asset_db_lookup}, asset.cache.hit: #{@asset_cache_hit}, parent_asset_created: #{@parent_asset_created}, duplicate_created_asset: #{@duplicate_created_asset} " if @logger
          @logger.info "Processing Job Metrics - timers: asset_type: #{@asset_type}, job.runtime: #{@job_duration / 1000} sec, cache_load_duration: #{@cache_load_duration} msec" if @logger
        end

        # record any custom counters
        @custom_counters.each do |k, v|
          @wavefront.counter(k, v, tags)
          @logger.debug("Custom Processing Job Metrics: asset_type: #{@asset_type}, metric: #{k}, count: #{v}") if @logger
        end
      end
    end
  end
end
