require_relative "account_resolver_helper"

module ContainersCp::Processing
  module MissingInstanceRepair
    # This module provides methods to find kubernetes_nodes with missing instance IDs (AWS, GCP and Azure),
    # extract the instance ID from the spec_provider_id and look up the instance. If found it will populate the
    # appropriate instance ID column for that node.
    #
    # This is used by the "instance repair job" to fix nodes that don't have this populated, typically due to the
    # node being short-lived, and the instance not being present in our DB at the initial collection time.
    #
    include ::ContainersCp::Processing::AccountResolverHelper

    @logger = Application["logger"]
    @wavefront = Application["wavefront"]

    LOOKBACK_TIME = ::Time.now - 14.days

    # each of these methods is called within a Multidb transaction, shard passed in for logging/metrics purposes.
    def repair_aws_instances(s)

      @logger.info "Looking for missing AWS instance IDs on shard #{s}\n"
      missing_aws_instance_nodes = ::KubernetesNode.where(cloud_provider_id: "AWS")
                                       .where('aws_instance_id IS NULL and spec_provider_id IS NOT NULL')
                                       .where("created_at > ?", LOOKBACK_TIME)

      @logger.info "Found #{missing_aws_instance_nodes.size} AWS nodes with missing instance_id on shard #{s}\n"
      found_aws_instance=0
      aws_instance_not_found=0
      aws_instance_mismatch=0
      aws_fargate_instance=0
      missing_aws_instance_nodes.each do |n|
        if provider_mismatch?(n.cloud_provider_id, n.spec_provider_id)
          @logger.debug "Cloud provider #{n.cloud_provider_id} doesn't match spec_provider_id: #{n.spec_provider_id} for node: #{n.id}"
          aws_instance_mismatch += 1
          next
        end
        instance_id = extract_instance_id_from_provider(n.cloud_provider_id, n.spec_provider_id)
        if instance_id.start_with?("fargate")
          @logger.debug "Skipping fargate instance: #{instance_id}"
          aws_fargate_instance += 1
          next
        end
        instance = ::AwsInstance.where(customer_id: n.customer_id, aws_account_id: n.aws_account_id, instance_id: instance_id).first
        if instance
          @logger.debug "Found instance: #{instance.id} : [#{instance.instance_id}], ins acct: #{instance.aws_account_id}, is_active: #{instance.is_active}, created_at: #{instance.created_at}, For node: #{n.name}, providerID: #{n.spec_provider_id}, node acct: #{n.aws_account_id}"
          n.aws_instance_id = instance.id
          n.save
          found_aws_instance += 1
        else
          @logger.debug "AWS Instance not found for ID: #{instance_id}"
          aws_instance_not_found += 1
        end
      end
      @wavefront.gauge("containers_cp.missing_instance_repair.aws_instance_found", found_aws_instance, shard: s)
      @wavefront.gauge("containers_cp.missing_instance_repair.aws_instance_not_found", aws_instance_not_found, shard: s)
      @wavefront.gauge("containers_cp.missing_instance_repair.aws_instance_mismatch", aws_instance_mismatch, shard: s)
      @wavefront.gauge("containers_cp.missing_instance_repair.aws_fargate_instance", aws_fargate_instance, shard: s)
      @logger.info "Completed processing shard: #{s}, missing AWS instances found: #{found_aws_instance}, not found:  #{aws_instance_not_found}"
    end

    def repair_gcp_instances(s)
      @logger.info "Looking for missing GCP instance IDs on shard #{s}\n"
      missing_gcp_instance_nodes = ::KubernetesNode.where(cloud_provider_id: "GCP")
                                       .where('gcp_compute_instance_id IS NULL and spec_provider_id IS NOT NULL')
                                       .where("created_at > ?", LOOKBACK_TIME)

      @logger.info "Found #{missing_gcp_instance_nodes.size} GCP nodes with missing gcp_compute_instance_id on shard #{s}\n"
      found_gcp_instance=0
      gcp_instance_not_found=0
      gcp_instance_mismatch=0
      missing_gcp_instance_nodes.each do |n|
        if provider_mismatch?(n.cloud_provider_id, n.spec_provider_id)
          @logger.debug "Cloud provider #{n.cloud_provider_id} doesn't match spec_provider_id: #{n.spec_provider_id} for node: #{n.id}"
          gcp_instance_mismatch += 1
          next
        end
        instance_id = extract_instance_id_from_provider(n.cloud_provider_id, n.spec_provider_id)
        instance = ::GcpComputeInstance
                       .where(customer_id: n.customer_id, gcp_compute_project_id:  n.gcp_compute_project_id, name: instance_id)
                       .first
        if instance
          @logger.debug "Found GCP instance: #{instance.id}:[#{instance.name}], is_active: #{instance.is_active}, created_at: #{instance.created_at}, For node: #{n.name}, providerID: #{n.spec_provider_id}"
          n.gcp_compute_instance_id = instance.id
          n.save
          found_gcp_instance += 1
        else
          @logger.debug "GCP Instance not found for ID: #{instance_id}"
          gcp_instance_not_found += 1
        end
      end
      @wavefront.gauge("containers_cp.missing_instance_repair.gcp_instance_found", found_gcp_instance, shard: s)
      @wavefront.gauge("containers_cp.missing_instance_repair.gcp_instance_not_found", gcp_instance_not_found, shard: s)
      @wavefront.gauge("containers_cp.missing_instance_repair.gcp_instance_mismatch", gcp_instance_mismatch, shard: s)
      @logger.info "Completed processing shard: #{s}, missing GCP instances found: #{found_gcp_instance}, not found:  #{gcp_instance_not_found}"
    end

    def repair_azure_instances(s)
      @logger.info "Looking for missing Azure instance IDs on shard #{s}\n"
      missing_azure_instance_nodes = ::KubernetesNode.where(cloud_provider_id: "AZURE")
                                         .where('azure_vm_id  IS NULL and spec_provider_id IS NOT NULL')
                                         .where("created_at > ?", LOOKBACK_TIME)

      @logger.info "Found #{missing_azure_instance_nodes.size} azure nodes with missing azure_vm_id  on shard #{s}\n"
      found_azure_instance=0
      azure_instance_not_found=0
      azure_instance_mismatch=0
      missing_azure_instance_nodes.each do |n|
        if provider_mismatch?(n.cloud_provider_id, n.spec_provider_id)
          @logger.debug "Cloud provider #{n.cloud_provider_id} doesn't match spec_provider_id: #{n.spec_provider_id} for node: #{n.id}"
          azure_instance_mismatch += 1
          next
        end
        instance_id = extract_instance_id_from_provider(n.cloud_provider_id, n.spec_provider_id)
        instance = ::AzureVm
                       .where(customer_id: n.customer_id, azure_subscription_id: n.azure_subscription_id, instance_id: instance_id)
                       .first
        if instance
          @logger.debug "Found azure VM instance: #{instance.id}:[#{instance.name}], is_active: #{instance.is_active}, created_at: #{instance.created_at}, For node: #{n.name}, providerID: #{n.spec_provider_id}"
          n.azure_vm_id = instance.id
          n.save
          found_azure_instance += 1
        else
          @logger.debug "Azure Instance not found for ID: #{instance_id}"
          azure_instance_not_found += 1
        end
      end
      @wavefront.gauge("containers_cp.missing_instance_repair.azure_instance_found", found_azure_instance, shard: s)
      @wavefront.gauge("containers_cp.missing_instance_repair.azure_instance_not_found", azure_instance_not_found, shard: s)
      @wavefront.gauge("containers_cp.missing_instance_repair.azure_instance_mismatch", azure_instance_mismatch, shard: s)
      @logger.info "Completed processing shard: #{s}, missing azure instances found: #{found_azure_instance}, not found:  #{azure_instance_not_found}"
    end

    def provider_mismatch? (provider_id, spec_provider_id)
      case provider_id
      when 'AWS'
        return true unless spec_provider_id.start_with?("aws")
      when 'GCP'
        return true unless spec_provider_id.start_with?("gce")
      when 'AZURE'
        return true unless spec_provider_id.start_with?("azure")
      end
      false
    end

  end
end
