module ContainersCp::Processing
  class TagUpdateHelper

    ID_SLICE_SIZE = 5000
    # This class provides methods to update both labels and annotations for kubernetes assets
    # Keeping the term "tag" around as the general term for both...
    def initialize(customer, asset_class, label_ids, annotation_ids, logger = nil)
      @customer = customer
      @asset_class = asset_class

      ::Multidb.use(@customer.read_shard) do
        @label_cache = {}
        label_ids.each_slice(ID_SLICE_SIZE) do |label_ids_slice|
          @asset_class.label_model.where(id: label_ids_slice).each_with_object({}) do |tag, h|
            @label_cache[[tag.key, tag.value]] = tag
          end
        end

        @annotation_cache = {}
        annotation_ids.each_slice(ID_SLICE_SIZE) do |annotation_ids_slice|
          @asset_class.annotation_model.where(id: annotation_ids).each_with_object({}) do |tag, h|
            @annotation_cache[[tag.key, tag.value]] = tag
          end
        end

        Application["logger"].debug "TagUpdateHelper.initialize, loaded #{@label_cache.size} labels into cache: , #{@annotation_cache.size} annotations."
      end
      @logger = logger
    end

    # Parses labels and annotations from the given dso and modifies the given asset if the
    # collected tags differ.
    def update_asset_tags(asset, dso_labels, dso_annotations)
      dso_labels ||= []
      dso_annotations ||= []

      # update labels for a terminated instance only if there were no labels or annotations already present
      if !asset.is_active?
        return if (dso_labels.blank? || !asset.labels.blank?) && (dso_annotations.blank? || !asset.annotations.blank?)
      end

      parsed_labels = parse_tags_from_dso(dso_labels)
      asset.labels = remove_emoji_characters(dso_labels.collect { |k, v| "#{k}: #{v}" }.sort.join(", "))
      if asset.labels_changed?
        begin
          perform_asset_labels_update(asset, parsed_labels)
        rescue => ex
          @logger.error "Exception updating asset labels: #{ex}-#{ex.message}" if @logger
        end
      end

      parsed_annotations = parse_tags_from_dso(dso_annotations)
      asset.annotations = remove_emoji_characters(dso_annotations.collect { |k, v| "#{k}: #{v}" }.sort.join(", "))
      if asset.annotations_changed?
        begin
          perform_asset_annotations_update(asset, parsed_annotations)
        rescue => ex
          @logger.error "Exception updating asset annotations: #{ex}-#{ex.message}" if @logger
        end
      end
    end

    private

    # Given a hash of dso_tags this will remove emoji characters from the key
    # and value of the hash
    def parse_tags_from_dso(dso_tags)
      dso_tags.map do |k, v|
        [remove_emoji_characters(k), remove_emoji_characters(v)]
      end.to_h
    end

    # Actually performs modification to the asset object based on a list of
    # parsed labels. It will modify the "labels" property and the tag relation of the given
    # asset.
    def perform_asset_labels_update(asset, parsed_tags)
      collected_tag_objects = []
      parsed_tags.each do |key, value|
        tag_object = lookup_label_by_key_and_value(key, value)
        if tag_object.nil?
          raise "Unable to create #{@asset_class.label_model.table_name} [#{[key, value]}] for customer #{@customer.id}"
        end

        collected_tag_objects << tag_object
      end
      tags_to_associate_with_asset = collected_tag_objects

      label_relation = "#{@asset_class.label_relation}=".to_sym
      asset.send label_relation, tags_to_associate_with_asset
    end

    # same method as above, for annotations
    def perform_asset_annotations_update(asset, parsed_tags)
      collected_tag_objects = []
      parsed_tags.each do |key, value|
        tag_object = lookup_annotation_by_key_and_value(key, value)
        if tag_object.nil?
          raise "Unable to create #{@asset_class.annotation_model.table_name} [#{[key, value]}] for customer #{@customer.id}"
        end
        collected_tag_objects << tag_object
      end
      tags_to_associate_with_asset = collected_tag_objects

      annotation_relation = "#{@asset_class.annotation_relation}=".to_sym
      asset.send annotation_relation, tags_to_associate_with_asset
    end

    # Looks up a tag object based on key and value. If it does not exist it will
    # create a new tag object. This method leverages a cache.
    def lookup_label_by_key_and_value(key, value)
      cache_key = [key, value]
      tag_obj = @label_cache[cache_key]
      if tag_obj.nil?
        begin
          tag_obj = @asset_class.label_model.find_or_create_by_customer_id_and_key_and_value_and_customer_tag(@customer.id, key, value)
        rescue ::ActiveRecord::RecordNotUnique
          tag_obj = @asset_class.label_model.find_by_customer_id_and_key_and_value_and_customer_tag(@customer.id, key, value)
        end

        @label_cache[cache_key] = tag_obj
      end

      tag_obj
    end

    def lookup_annotation_by_key_and_value(key, value)
      cache_key = [key, value]
      tag_obj = @annotation_cache[cache_key]
      if tag_obj.nil?
        begin
          tag_obj = @asset_class.annotation_model.find_or_create_by_customer_id_and_key_and_value_and_customer_tag(@customer.id, key, value)
        rescue ::ActiveRecord::RecordNotUnique
          tag_obj = @asset_class.annotation_model.find_by_customer_id_and_key_and_value_and_customer_tag(@customer.id, key, value)
        end
        @annotation_cache[cache_key] = tag_obj
      end

      tag_obj
    end

    def remove_emoji_characters(item)
      return item if item.blank?
      return item if item.each_char.all? { |c| c.bytes.count < 4 }
      item.each_char.select { |c| c.bytes.count < 4 }.join("")
    end
  end
end
