# frozen_string_literal: true

module ContainersCp
  class ShardLookup
    include Import["logger", "alert", "cache.redis", "cache.common_assets"]

    # for some reason "15.minutes" was throwing an error here
    FIFTEEN_MIN_TTL = 15 * 60
    def initialize(**kwargs)
      @stores = ::Hash.new { |h, k| h[k] = {} }
      super
    end

    def main_shard(customer)
      "db#{lookup_shard_id(customer)}"
    end

    def read_shard(customer)
      "read#{lookup_shard_id(customer)}"
    end

    # Flush cached shard assignments from redis
    # This makes it so we don't need to perform any manual steps after a shard split/change
    def flush_redis_cache
      namespace = redis.client.options[:namespace]
      matching = "'#{namespace}:shard_assignments:*'"
      logger.warn "Flushing shard assignment cache matching #{matching}"
      redis.eval "return redis.call('del', unpack(redis.call('keys', #{matching})))"
    rescue => ex
      # This error happens if no keys are found. (This can be common in dev)
      return nil if ex.message.include? "Wrong number of args calling Redis command From Lua script"
      logger.error "Failed to flush shard assignments: #{ex.message}"
      alert.exception ex
      nil
    end

    private

    def lookup_shard_id(customer)
      name = ::DataStore::MAIN
      shard_id = @stores[name][customer.id] || fetch_redis(name, customer) || fetch_database(name, customer)
      raise "Shard assignment not found" unless shard_id
      shard_id
    end

    def fetch_redis(name, customer)
      key = shard_assignment_cache_key customer
      shard_id = redis.from_hash_field key, name
      return nil unless shard_id
      @stores[name][customer.id] = shard_id
      shard_id
    rescue => ex
      logger.error "Failed to fetch shard assignment from redis cache: #{ex.class}"
      nil
    end

    def store_redis(name, customer, shard_id)
      key = shard_assignment_cache_key customer
      redis.add_to_hash key, name, shard_id, FIFTEEN_MIN_TTL
    rescue => ex
      logger.error "Failed to store shard assignment from redis cache: #{ex.class}"
    end

    def shard_assignment_cache_key(customer)
      "shard_assignments:#{customer.id}"
    end

    def fetch_database(name, customer)
      ::Multidb.use(:master) do
        ds = common_assets.data_store(name)
        sa = ::ShardAssignment.where(customer_id: customer.id, data_store_id: ds.id).first
        return nil unless sa
        @stores[name][customer.id] = sa.shard_id
        store_redis name, customer, sa.shard_id
        sa.shard_id
      end
    end
  end
end
