module ContainersCp::Workers
  #
  # This worker class handles a couple of scenario where we need to update all assets belonging to a specified cluster.
  # These are specified by the "action" job argument.
  #
  # 1) Account Update - when we first identify the owner account for a cluster, we trigger this job to propagate
  #    the owner account info to all assets under this cluster, as some may have been added without account info.
  #
  # 2) Cluster Remove (deactivation) - When the Event Dispatcher receives a "cluster_remove" event type, this job
  #    is triggered to deactivate the cluster itself, then find and deactivate all child assets under that cluster.
  #
  class AssetUpdateWorker
    include Sidekiq::Worker
    include Import["logger", "alert", "wavefront"]

    sidekiq_options queue: :processing_assets_containers

    # Set on context to indicate which type of update we're doing
    ACTION_UPDATE_ACCOUNT_INFO = "acct_update"
    ACTION_CLUSTER_REMOVE = "cluster_remove"

    # retry a couple of times, but don't worry too much - the account IDs will get updated the next time we
    # receive a feed for the assets for that cluster anyway.
    MAX_UPDATE_RETRIES = 3

    def perform(context)
      # default to account update for backward-compatibility on rollout
      @action = context["action"] || ACTION_UPDATE_ACCOUNT_INFO

      # use consistent timestamp for inactivated_at time
      @db_update_timestamp = ::Time.now

      logger.debug "Account Update job - action: #{@action}, args: #{context}"
      if validate_context(context)
        customer_id = context["customer_id"]
        cluster_id = context["parent_cluster_id"]

        customer = ::Multidb.use(:master) { ::Customer.find(customer_id) }
        raise ::Platform::CustomerNotFoundError unless customer

        # for a cluster remove, deactivate the cluster first, then update all child assets
        if @action == ACTION_CLUSTER_REMOVE
          cluster_id = deactivate_cluster(customer, context["cluster_id"])
        end

        k8s_asset_types = get_k8s_asset_models
        logger.debug "Found #{k8s_asset_types.size} asset types to update."
        k8s_asset_types.each do |model_class|
          update_assets(customer, model_class, context, cluster_id)
        end
      else
        logger.error("Missing required properties for account update job, context: #{context}")
      end
    end

    def get_k8s_asset_models
      asset_types_to_update = []
      KubernetesAssetBase.descendants.each do |desc|
        # don't include nodes if we're updating account info (this is triggered by nodes processing)
        next if desc.name == "KubernetesNode" && @action == ACTION_UPDATE_ACCOUNT_INFO
        asset_types_to_update << desc if is_valid_model_type(desc)
      end
      asset_types_to_update
    end

    # ensure that the model type has all of the required account ID fields to update
    def is_valid_model_type(model_class)
      return false unless model_class.column_names.include?("cloud_provider_id")
      return false unless model_class.column_names.include?("cloud_account_id")
      return false unless model_class.column_names.include?("aws_account_id")
      return false unless model_class.column_names.include?("gcp_compute_project_id")
      return false unless model_class.column_names.include?("azure_subscription_id")
      true
    end

    def update_assets(customer, model_class, context, cluster_id)
      return unless cluster_id

      logger.debug "update_assets, model_class: #{model_class.name}, @customer.id: #{customer.id}, shard: #{customer.data_shard}, cluster_id: #{cluster_id}"
      retries = 0
      begin
        ::Multidb.use(customer.data_shard) do
          assets = model_class.active.where(customer_id: customer.id, containers_cluster_id: cluster_id)
          logger.debug "Found #{assets.size} assets of type: #{model_class.name} to update"
          assets.each do |asset|
            if @action == ACTION_UPDATE_ACCOUNT_INFO
              set_asset_account_info(asset, context)
            else
              deactivate_asset(asset)
            end
            asset.save
          end
          logger.debug "Completed asset update for type #{model_class.name}"
        end
      rescue => ex
        logger.warn("DB exception updating asset type #{model_class.name}: #{ex}-#{ex.message}, retrying...")
        if retries <= MAX_UPDATE_RETRIES
          retries += 1
          sleep 2**retries
          retry
        else
          msg = "Giving up on the DB account info update for asset type: #{model_class.name} after #{retries} retries. Error: #{ex.message}"
          logger.error(msg)
          raise msg
        end
      end
    end

    def set_asset_account_info(asset, context)
      asset.cloud_provider_id = context["cloud_provider"]
      asset.cloud_account_id = context["cloud_account_id"]
      asset.aws_account_id = context["aws_account_id"] if context["aws_account_id"]
      asset.gcp_compute_project_id = context["gcp_compute_project_id"] if context["gcp_compute_project_id"]
      asset.azure_subscription_id = context["azure_subscription_id"] if context["azure_subscription_id"]
    end

    def deactivate_cluster(customer, cluster_id)
      begin
        logger.debug "deactivating cluster - customer_id: #{customer.id}, cluster_id: #{cluster_id}"
        ::Multidb.use(customer.data_shard) do
          cluster = ::ContainersCluster.find(cluster_id)
          if cluster
            logger.debug "found cluster #{cluster_id}, cluster name: #{cluster.cluster_name}"
            deactivate_asset(cluster)
            cluster.save
          else
            logger.warn "Cluster to deactivate not found: #{cluster_id} for customer: #{customer.id}"
          end
        end
      rescue => ex
        logger.warn("DB exception deactivating cluster #{cluster_id}: #{ex}-#{ex.message}, retrying...")
        if retries <= MAX_UPDATE_RETRIES * 2
          retries += 1
          sleep 2**retries
          retry
        else
          msg = "Giving up on deactivating cluster #{cluster_id} after #{retries} retries. Error: #{ex.message}"
          logger.error(msg)
          raise msg
        end
      end
      cluster_id
    end

    def deactivate_asset(asset)
      asset.is_active = false
      asset.inactivated_at = @db_update_timestamp
    end

    # ensure the job context has the info required to do the account info updates
    def validate_context(context)
      # customer_id required by both actions
      return false unless context["customer_id"]
      # account info update requires additional params
      if @action == ACTION_UPDATE_ACCOUNT_INFO
        return false unless context["parent_cluster_id"]
        return false unless context["cloud_account_id"]
        case context["cloud_provider"]
        when PROVIDER_AWS
          return false unless context["aws_account_id"]
        when PROVIDER_GCP
          return false unless context["gcp_compute_project_id"]
        when PROVIDER_AZURE
          return false unless context["azure_subscription_id"]
        else
          return false
        end
      elsif @action == ACTION_CLUSTER_REMOVE
        return false unless context["cluster_id"]
      else
        return false
      end
      true
    end
  end
end
